/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.security.idm.ldap;

//core java
import java.security.Principal;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Properties;
import java.text.MessageFormat;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import org.apache.log4j.Logger;
import org.jboss.security.idm.Identity;
import org.jboss.security.idm.LoginContext;
import org.jboss.security.idm.Role;
import org.jboss.security.idm.IdentityException;
import org.jboss.security.idm.ProvisioningProvider;
import org.jboss.security.util.Tools;



/**
 * LDAPIdentityProvider is a built-in implementation of a ProvisioningProvider/LoginProvider that accessed data stored in an LDAP repository<br/><br/>
 * 
 * This implementation supports the following standard LDAP schemas:<br/>
 * User Identity Information stored in InetOrgPerson schema<br/>
 * Role Information stored in GroupOfUniqueNames schema<br/>
 * 
 * At this time, this implementation is tested to work with OpenLDAP and RedHat Directory Server.
 * 
 * 
 * @author Sohil Shah - sohil.shah@jboss.com - May 30, 2006
 *
 */
public class LDAPIdentityProvider implements ProvisioningProvider
{
    /**
     * 
     */
    private static final Logger log = Logger.getLogger(LDAPIdentityProvider.class);
    
    /**
     * sql statements for identity
     */
    private static final String readSQL = "select cn,sn,userPassword,givenName,displayName,o,employeeType,title,postalAddress,mail," +
    "telephoneNumber from cn={0},ou={1}";
    private static final String insertSQL = "insert into cn,ou={0}"+
    "(objectclass,cn,sn,userPassword,givenName,displayName,o,employeeType,title,{1},{3},{5}" +             
    ") " +
    "VALUES " +
    "(" +
    "?,?,?,?,?,?,?,?,?,{2},{4},{6}" +
    ")";
    private static final String removeSQL = "delete from ou={0} where cn=?";
    
    /**
     * sql statements for roles
     */
    private static final String readRoleSQL = "select cn,uniqueMember from cn={0},ou={1}";
    private static final String readAllRolesSQL = "select cn,uniqueMember from ou={0} where objectclass=groupOfUniqueNames";
    private static final String insertRoleSQL = "insert into cn,ou={0}"+
    "(objectclass,cn,{1}" +             
    ") " +
    "VALUES " +
    "(" +
    "?,?,{2}" +
    ")";    
    private static final String removeRoleSQL = "delete from ou={0} where cn=?";
    
    
    /**
     * 
     */
    protected String id = null;
    protected String connectionURL = null;
    protected String username = null;
    protected String password = null;
    protected String identityOu = null;
    protected String roleOu = null;
    
    
    /**
     * 
     *
     */
    public LDAPIdentityProvider()
    {        
    }
    
    /**
     * 
     * @author Sohil Shah - sohil.shah@jboss.com
     *
     */
    private static class ParamIndex
    {
        private int index = 1;
        private ParamIndex()
        {           
        }
        int getIndex()
        {
            return index++;
        }
    }
                
                
    /**
     * 
     * @param values
     * @param token
     * @return
     */
    protected String getMultiValueColumns(Collection values,String token)
    {
        return getMultiValueColumns(values,token,",");
    }
    
    /**
     * 
     * @param values
     * @param token
     * @param separator
     * @return
     */
    protected String getMultiValueColumns(Collection values,String token,String separator)
    {
        String result = "";
        if(!Tools.isEmpty(values))
        {
            StringBuffer buffer = new StringBuffer();
            for(Iterator itr=values.iterator();itr.hasNext();)
            {
                itr.next();
                buffer.append(token);
                if(itr.hasNext())
                {
                    buffer.append(separator);
                }
            }
            result = buffer.toString();
        }
        return result;
    }
    
    /**
     * 
     * @param value
     * @return
     */
    protected Collection getMultiValues(String value)
    {
        Collection values = new ArrayList();
        
        if(!Tools.isEmpty(value))
        {           
            if(value.indexOf('[')!=-1)
            {
                int startIndex = 0;
                while((startIndex=value.indexOf('[',startIndex))!=-1)
                {
                    int endIndex = value.indexOf(']',startIndex);
                    String cour = value.substring(startIndex+1,endIndex);
                    if(!cour.equalsIgnoreCase(Identity.NULL))
                    {
                        values.add(cour);
                    }
                    startIndex = endIndex;
                    if(startIndex>=value.length())
                    {
                        break;
                    }
                }
            }
            else
            {
                values.add(value);
            }
        }
        
        return values;
    }
    
    protected Identity readIdentity(String sql,String username) throws SQLException
    {
        Identity identity = null;       
        try
        {           
            Connection connection = this.getConnection();
            try
            {                       
                String select = MessageFormat.format(sql,new Object[]{username.trim(),this.identityOu});
                Statement st = connection.createStatement();                
                ResultSet rs = st.executeQuery(select);
                try
                {
                   Collection identities = this.readIdentites(rs);
                   
                   if(!Tools.isEmpty(identities))
                   {
                       //this should return only one identity
                       identity = (Identity)identities.iterator().next();
                   }
                }
                finally
                {
                    if(rs!=null)
                    {
                        rs.close();
                    }
                    if(st!=null)
                    {
                        st.close();
                    }
                }
            }
            finally
            {
                if(connection!=null)
                {
                    connection.close();
                }
            }
        }
        catch(SQLException sqe)
        {
            if(sqe.toString().indexOf("No Such Object")!=-1)
            {
                identity = null;
            }
            else
            {
                throw new IdentityException(sqe);
            }
        }       
        return identity;
    }
            
    /**
     * identity population method to be used by others...this is refactored code
     * @param rs
     * @return
     */
    protected Collection readIdentites(ResultSet rs) throws SQLException
    {
        Collection identities = new ArrayList();
        while(rs.next())
        {
            Identity identity = new Identity();
            
            String cn = rs.getString("cn");
            String cour = rs.getString("sn");
            boolean active = (new Boolean(cour)).booleanValue();
            String pass  = rs.getString("userPassword");                       
            String givenName = rs.getString("givenName");
            String companyName = rs.getString("o");
            String companyRole = rs.getString("employeeType");
            String companyTitle = rs.getString("title");
            String postalAddress = rs.getString("postalAddress");
            
            //populate the identity
            identity = new Identity();
            identity.setUserName(cn);
            identity.setPassword(pass.getBytes());
            identity.setActive(active);                       
            
            //optional fields                       
            identity.setFullName(givenName);
            
            //company fields                       
            identity.setCompanyName(companyName);
            identity.setCompanyRole(companyRole);
            identity.setCompanyTitle(companyTitle);
            
            //address fields                         
            Collection addressValues = this.getMultiValues(postalAddress);
            int counter=0;
            for(Iterator itr=addressValues.iterator();itr.hasNext();)
            {
                String addressItem = (String)itr.next();
                switch(counter)
                {
                    case 0:
                        if(addressItem.indexOf('=')==-1)
                        {
                            identity.getCompanyAddress().setStreet(addressItem);
                        }
                    break;
                    
                    case 1:
                        if(addressItem.indexOf('=')==-1)
                        {
                            identity.getCompanyAddress().setCity(addressItem);
                        }
                    break;
                    
                    case 2:
                        if(addressItem.indexOf('=')==-1)
                        {
                            identity.getCompanyAddress().setState(addressItem);
                        }
                    break;
                    
                    case 3:
                        if(addressItem.indexOf('=')==-1)
                        {
                            identity.getCompanyAddress().setPostalCode(addressItem);
                        }
                    break;
                    
                    case 4:
                        if(addressItem.indexOf('=')==-1)
                        {
                            identity.getCompanyAddress().setCountry(addressItem);
                        }
                    break;
                }
                counter++;
            }
            
            //grab the communication related information
            String emails = rs.getString("mail");
            String phoneNumbers = rs.getString("telephoneNumber");
            Collection allEmails = this.getMultiValues(emails);
            Collection allPhoneNumbers = this.getMultiValues(phoneNumbers);                       
            identity.setPhoneNumbers(getMultiValues(phoneNumbers));
            if(!Tools.isEmpty(allEmails))
            {
                Object[] emailArray = allEmails.toArray();
                identity.setEmail((String)emailArray[0]);
                for(int i=1;i<emailArray.length;i++)
                {
                    identity.addEmail((String)emailArray[i]);
                }
            }
            
            identities.add(identity);
        }
        return identities;
    }
    
    /**
     * 
     * @param identity
     * @throws Exception
     */
    protected void storeIdentity(Identity identity,boolean mustEncryptPassword) throws Exception
    {
        Connection connection = this.getConnection();
        ParamIndex p = new ParamIndex();
        Identity jbIdentity = identity;
        try
        {
            //get email multivalues
            Collection emails = new ArrayList();
            emails.add(identity.getEmail());
            emails.addAll(identity.getOtherEmails());
            if(Tools.isEmpty(emails))
            {
                emails.add(Identity.NULL);
            }
            String emailColumns = this.getMultiValueColumns(emails,"mail");
            String emailValues = this.getMultiValueColumns(emails,"?");
            
            //get phone multivalues
            Collection phones = new ArrayList();
            phones.addAll(identity.getPhoneNumbers());
            if(Tools.isEmpty(phones))
            {
                phones.add(Identity.NULL);
            }
            String phoneColumns = this.getMultiValueColumns(phones,"telephoneNumber");
            String phoneValues = this.getMultiValueColumns(phones,"?");
            
            //get postalAddress multivalues
            Collection postalAddress = new ArrayList();
            if(!Tools.isEmpty(identity.getCompanyAddress().getStreet()))
            {
                postalAddress.add(identity.getCompanyAddress().getStreet());
            }
            else
            {
                postalAddress.add("street="+Identity.NULL);
            }
            if(!Tools.isEmpty(identity.getCompanyAddress().getCity()))
            {
                postalAddress.add(identity.getCompanyAddress().getCity());
            }
            else
            {
                postalAddress.add("city="+Identity.NULL);
            }
            if(!Tools.isEmpty(identity.getCompanyAddress().getState()))
            {
                postalAddress.add(identity.getCompanyAddress().getState());
            }
            else
            {
                postalAddress.add("state="+Identity.NULL);
            }
            if(!Tools.isEmpty(identity.getCompanyAddress().getPostalCode()))
            {
                postalAddress.add(identity.getCompanyAddress().getPostalCode());
            }
            else
            {
                postalAddress.add("postalCode="+Identity.NULL);
            }
            if(!Tools.isEmpty(identity.getCompanyAddress().getCountry()))
            {
                postalAddress.add(identity.getCompanyAddress().getCountry());
            }
            else
            {
                postalAddress.add("country="+Identity.NULL);
            }
            String postalAddressColumns = this.getMultiValueColumns(postalAddress,"postalAddress");
            String postalAddressValues = this.getMultiValueColumns(postalAddress,"?");
            
            
            String insert = MessageFormat.format(insertSQL,new Object[]{this.identityOu,emailColumns,emailValues,phoneColumns,phoneValues,
            postalAddressColumns,postalAddressValues});
                        
            PreparedStatement st = connection.prepareStatement(insert);
            
            //generic identity data
            String password = new String(identity.getPassword());            
            if(mustEncryptPassword)
            {                
                String encodedPassword = Tools.getEncodedPassword(password,"MD5");
                password = encodedPassword;
            }           
            
            //these are the generic identity fields and are required from a JBoss Corp perspective
            st.setString(p.getIndex(),"inetOrgPerson");
            st.setString(p.getIndex(),identity.getUserName());
            st.setBoolean(p.getIndex(),identity.isActive());                
            st.setString(p.getIndex(),password); 
            
            if(!Tools.isEmpty(jbIdentity.getFullName()))
            {
                st.setString(p.getIndex(),jbIdentity.getFullName());
            }
            else
            {
                st.setString(p.getIndex(),"null");
            }
            
            if(!Tools.isEmpty(jbIdentity.getFullName()))
            {
                st.setString(p.getIndex(),jbIdentity.getFullName());
            }
            else
            {
                st.setString(p.getIndex(),"null");
            }
            
            if(!Tools.isEmpty(jbIdentity.getCompanyName()))
            {
                st.setString(p.getIndex(),jbIdentity.getCompanyName());
            }
            else
            {
                st.setString(p.getIndex(),"null");
            }
            
            if(!Tools.isEmpty(jbIdentity.getCompanyRole()))
            {
                st.setString(p.getIndex(),jbIdentity.getCompanyRole());
            }
            else
            {
                st.setString(p.getIndex(),"null");
            }
            
            if(!Tools.isEmpty(jbIdentity.getCompanyTitle()))
            {
                st.setString(p.getIndex(),jbIdentity.getCompanyTitle());
            }
            else
            {
                st.setString(p.getIndex(),"null");
            }
            
            
            //insert emails
            for(Iterator itr=emails.iterator();itr.hasNext();)
            {
                String email = (String)itr.next();
                st.setString(p.getIndex(),email);
            }
            //insert phones
            for(Iterator itr=phones.iterator();itr.hasNext();)
            {
                String phone = (String)itr.next();
                st.setString(p.getIndex(),phone);
            }
            //insert postal address
            for(Iterator itr=postalAddress.iterator();itr.hasNext();)
            {
                String addressItem = (String)itr.next();
                st.setString(p.getIndex(),addressItem);
            }
                                                            
            try
            {
               st.executeUpdate();
            }
            finally
            {                
                if(st!=null)
                {
                    st.close();
                }
            }
        }
        finally
        {
            if(connection!=null)
            {
                connection.close();
            }
        }
    }
    
    /**
     * 
     * @param rs
     * @return
     * @throws Exception
     */
    protected Collection readRoles(ResultSet rs) throws SQLException
    {
        Collection roles = new ArrayList();
        
        while(rs.next())
        {
            Role role = new Role();
            String cn = rs.getString("cn");
            
            
            //grab the member related information
            String members = rs.getString("uniqueMember");
            Collection identities = this.getMultiValues(members);
            Collection parsed = new ArrayList();
            for(Iterator itr=identities.iterator();itr.hasNext();)
            {
                String member = (String)itr.next();
                if(!member.equals(Identity.NULL))
                {
                    int startIndex = 0;
                    if(member.indexOf('=')!=-1)
                    {
                        startIndex = member.indexOf('=') + 1;
                    }
                    parsed.add(member.substring(startIndex));
                }
            }
            
            role.setName(cn);
            role.setIdentities(parsed);
            
            roles.add(role);
        }
              
        return roles;
    }
    
    /**
     * 
     * @param identity
     * @throws Exception
     */
    protected void storeRole(Role role) throws Exception
    {
        Connection connection = this.getConnection();
        ParamIndex p = new ParamIndex();
        try
        {                        
            //get members
            Collection members = new ArrayList();
            members.addAll(role.getIdentities());
            if(Tools.isEmpty(members))
            {
                members.add(Identity.NULL);
            }
            String memberColumns = this.getMultiValueColumns(members,"uniqueMember");
            String memberValues = this.getMultiValueColumns(members,"?");
            
                        
            
            String insert = MessageFormat.format(insertRoleSQL,new Object[]{this.roleOu,memberColumns,memberValues});
                        
            PreparedStatement st = connection.prepareStatement(insert);
            
                       
            //these are the generic identity fields and are required from a JBoss Corp perspective
            st.setString(p.getIndex(),"groupOfUniqueNames");
            st.setString(p.getIndex(),role.getName());
                                  
            //insert members
            for(Iterator itr=members.iterator();itr.hasNext();)
            {
                String member = (String)itr.next();
                st.setString(p.getIndex(),"cn="+member);
            }
            
                                                            
            try
            {
               st.executeUpdate();
            }
            finally
            {                
                if(st!=null)
                {
                    st.close();
                }
            }
        }
        finally
        {
            if(connection!=null)
            {
                connection.close();
            }
        }
    }
    
    /**
     * 
     * @param identity
     * @throws IdentityException
     */
    protected void validateIdentity(Identity identity) throws IdentityException
    {
        if(
                identity == null ||
                Tools.isEmpty(identity.getUserName()) ||
                Tools.isEmpty(identity.getEmail()) ||
                identity.getPassword() == null
        )
        {
            throw new IdentityException("Identity Validation Failed...Please make sure the required fields are not empty");
        }
    }
    
    /**
     * 
     * @param role
     * @throws IdentityException
     */
    protected void validateRole(Role role) throws IdentityException
    {
        if(
                role == null ||
                Tools.isEmpty(role.getName())
        )
        {
            throw new IdentityException("Role Validation Failed...Please make sure the required fields are not empty");
        }
    }
    
    /**
     * 
     */
    protected Connection getConnection() throws SQLException
    {
        Connection connection = null;
        String ldapConnectString = this.connectionURL;
        String ldapUserName = this.username;
        String password = this.password;
        connection = DriverManager.getConnection(ldapConnectString,ldapUserName,password);
        return connection;
    }
    //identiyprovider implementation-------------------------------------------------------------------------------------------------------------
    /**
     * Returns an id that uniquely identifies the LoginProvider plugged into the JBoss SSO system. 
     * It should be of the form: si:<a unique string such as vendor name, etc. eg. jboss>:<optional parameter>
     * 
     * @return  id of the LoginProvider
     * @throws IdentityException
     */
    public String getId() 
    {
        return id;
    }

    /**
     * 
     * @param id
     */
    public void setId(String id) 
    {
        this.id = id;
    }
    
    /**
     * Sets any properties like LDAP URLs, JDBC configuration, etc needed to configure this particular implementation
     * of the LoginProvider
     * 
     * @param properties
     * @throws IdentityException
     */
    public void setProperties(Properties properties) throws IdentityException
    {
        //load the JDBC-LDAP driver
        try
        {
            Class.forName("com.octetstring.jdbcLdap.sql.JdbcLdapDriver");
        }        
        catch(ClassNotFoundException cne)
        {
            throw new RuntimeException(cne);
        }
        
        this.connectionURL = properties.getProperty("connectionURL");
        this.username = properties.getProperty("username");
        this.password = properties.getProperty("password");
        this.identityOu = properties.getProperty("identityOu");
        this.roleOu = properties.getProperty("roleOu");
        this.restoreIdentities();
        this.restoreRoles();
    }

    /**
     * Activates a user account
     * 
     * @param identity data of the user account that should be activated
     * @throws IdentityException
     */    
    public void activate(Identity identity) throws IdentityException 
    {
        this.activate(identity.getUserName());        
    }

    /**
     * Activates a user account
     * 
     * @param username of the user account that should be activated
     * @throws IdentityException
     */
    public void activate(String username) throws IdentityException 
    {
        Identity identity = this.read(username);
        if(!identity.isActive())
        {
            identity.setActive(true);
            this.update(identity);
        }        
    }

    /**
     * Deactivates a user account
     * 
     * @param identity data of the user account that should be deactivated
     * @throws IdentityException
     */
    public void deactivate(Identity identity) throws IdentityException 
    {
        this.deactivate(identity.getUserName());        
    }

    /**
     * Deactivates a user account
     * 
     * @param username of the user account that should be deactivated
     * @throws IdentityException
     */
    public void deactivate(String username) throws IdentityException 
    {
        Identity identity = this.read(username);
        if(identity.isActive())
        {
            identity.setActive(false);
            this.update(identity);
        }        
    }

    /**
     * Checks if the specified Principal exists in the system anymore or not
     * 
     * @param principal
     * @return
     * @throws IdentityException
     */
    public boolean exists(Principal principal) throws IdentityException 
    {
        return this.exists(principal.getName());
    }

    /**
     * Checks if the specified username exists in the system anymore or not
     * 
     * @param username
     * @return
     * @throws IdentityException
     */
    public boolean exists(String username) throws IdentityException 
    {
        boolean exists = false; 
        try
        {           
            Connection connection = this.getConnection();
            try
            {                       
                String select = "select cn "+
                "from cn="+username.trim()+",ou="+this.identityOu;
                Statement st = connection.createStatement();                
                ResultSet rs = st.executeQuery(select);
                try
                {
                   exists = rs.next();
                }
                finally
                {
                    if(rs!=null)
                    {
                        rs.close();
                    }
                    if(st!=null)
                    {
                        st.close();
                    }
                }
            }
            finally
            {
                if(connection!=null)
                {
                    connection.close();
                }
            }
        }
        catch(SQLException sqe)
        {
            if(sqe.toString().indexOf("No Such Object")!=-1)
            {
                exists = false;
            }
            else
            {
                throw new IdentityException(sqe);
            }
        }
        return exists;
    }

    /**
     * Checks if the specified Principal should successfully authenticated using the specified password
     * 
     * @param principal
     * @param password
     * @return
     * @throws IdentityException
     */
    public boolean login(Principal principal, byte[] password) throws IdentityException 
    {
        return this.login(principal.getName(),password);
    }

    /**
     * Checks if the specified username should be successfully authenticated using the specified password
     * 
     * @param username
     * @param password
     * @return
     * @throws IdentityException
     */
    public boolean login(String username, byte[] password) throws IdentityException 
    {
        try
        {
            boolean loginSuccess = false;
            
            Identity identity = this.read(username);
            
            String storedPassword = new String(identity.getPassword());
            String inputPassword = new String(Tools.getEncodedPassword(new String(password),"MD5"));
            if(storedPassword.equals(inputPassword) && identity.isActive())
            {
                loginSuccess = true;
            }
            
            return loginSuccess;
        }
        catch(Exception e)
        {
            throw new IdentityException(e);
        }
    }

    /**
     * Returns the Identity data for the specified Principal
     * 
     * @param principal
     * @return Identity data for the Principal
     * @throws IdentityException
     */
    public Identity read(Principal principal) throws IdentityException 
    {
        return this.read(principal.getName());
    }

    /**
     * Returns the Identity data for the specified username
     * 
     * @param username
     * @return Identity data for the username
     * @throws IdentityException
     */
    public Identity read(String username) throws IdentityException 
    {
        try
        {
            Identity identity = this.readIdentity(readSQL,username);                      
            return identity;
        }
        catch(SQLException sqe)
        {
            throw new IdentityException(sqe);
        }
    }

    /**
     * Registers a new user into the system
     * 
     * @param identity
     * @throws IdentityException
     */
    public void register(Identity identity) throws IdentityException 
    {
        //perform corp wide validation of the identity    
        this.validateIdentity(identity);
        try
        {
            this.storeIdentity(identity,true);
        }
        catch(Exception e)
        {
            e.printStackTrace();
            throw new IdentityException(e);
        }        
    }

    /**
     * Removes a user from the system
     * 
     * @param identity
     * @throws IdentityException
     */
    public void remove(Identity identity) throws IdentityException 
    {
        this.remove(identity.getUserName());        
    }

    /**
     * Removes a user from the system
     * 
     * @param username
     * @throws IdentityException
     */
    public void remove(String username) throws IdentityException 
    {
        try
        {
            Connection connection = this.getConnection();
            try
            {                    
                String delete = MessageFormat.format(removeSQL,new Object[]{this.identityOu});
                
                PreparedStatement st = connection.prepareStatement(delete);
                st.setString(1,username);                        
                try
                {
                   st.executeUpdate();
                }
                finally
                {                
                    if(st!=null)
                    {
                        st.close();
                    }
                }
            }
            finally
            {
                if(connection!=null)
                {
                    connection.close();
                }
            }
        }
        catch(SQLException sqe)
        {
            throw new IdentityException(sqe);
        }        
    }

    /**
     * Updates the user profile data
     * 
     * @param identity
     * @throws IdentityException
     */
    public void update(Identity identity) throws IdentityException 
    { 
        //perform corp wide validation of the identity    
        this.validateIdentity(identity);
        
        //cleanup any temporary backed up identities
        this.cleanupBackup(identity.getUserName());
        
        //make sure that there aren't any backup files existing for this identity
        //if they do, the identity data for this user needs to be cleaned up
        //while it is in this state, there should be no further changes made to this
        //information
        if(this.isIdentityCorrupted(identity.getUserName()))
        {
            throw new IdentityException(identity.getUserName()+" is corrupted. Please cleanup before making any updates on this identity");
        }
                        
        //make a temp copy of the entire identity that needs to be updated..
        //this copy is used for restoration if there are issues with this update
        Identity temp = this.read(identity.getUserName());
        String tempUserName = temp.getUserName() + ".bak" + "." + System.currentTimeMillis();
        temp.setUserName(tempUserName);             
        this.modifyInfo(temp); //write operation
        
        //now perform the actual update of the identity in the LDAP server
        try
        {           
            this.remove(identity); //write operation    
            this.modifyInfo(identity); //write operation
        }
        catch(Exception e)
        {
            //restore the original identity from the backed up identity
            //kind of a simulation of rollback          
            if(!this.exists(identity.getUserName()))
            {
                //this user's identity has been deleted, we need to restore it from the backed up copy
                Identity oldIdentity = this.read(tempUserName);
                oldIdentity.setUserName(identity.getUserName());
                this.modifyInfo(oldIdentity); //write operation
            }
            throw new IdentityException(e);
        }
        finally
        {               
            this.cleanupBackup(identity.getUserName());
        }
    } 
    
    /**
     * Resets the password of the specified user
     * 
     * @param username
     * @param password
     * @throws IdentityException
     */
    public void resetPassword(String username,byte[] password) throws IdentityException
    {
        Identity identity = this.read(username);
                
        //cleanup any temporary backed up identities
        this.cleanupBackup(identity.getUserName());
        
        //make sure that there aren't any backup files existing for this identity
        //if they do, the identity data for this user needs to be cleaned up
        //while it is in this state, there should be no further changes made to this
        //information
        if(this.isIdentityCorrupted(identity.getUserName()))
        {
            throw new IdentityException(identity.getUserName()+" is corrupted. Please cleanup before making any updates on this identity");
        }
                        
        //make a temp copy of the entire identity that needs to be updated..
        //this copy is used for restoration if there are issues with this update
        Identity temp = this.read(identity.getUserName());
        String tempUserName = temp.getUserName() + ".bak" + "." + System.currentTimeMillis();
        temp.setUserName(tempUserName);             
        this.modifyInfo(temp); //write operation
        
        //now perform the actual update of the identity in the LDAP server
        try
        {           
            this.remove(identity); //write operation
            
            identity.setPassword(password);
            this.storeIdentity(identity,true); //write operation
        }
        catch(Exception e)
        {
            //restore the original identity from the backed up identity
            //kind of a simulation of rollback          
            if(!this.exists(identity.getUserName()))
            {
                //this user's identity has been deleted, we need to restore it from the backed up copy
                Identity oldIdentity = this.read(tempUserName);
                oldIdentity.setUserName(identity.getUserName());
                this.modifyInfo(oldIdentity); //write operation
            }
            throw new IdentityException(e);
        }
        finally
        {               
            this.cleanupBackup(identity.getUserName());
        }
    }
    
    //-------------role related operations-----------------------------------------------------------------------------
    /**
     * Checks if the specified role exists in the system
     * 
     * @param role
     * @return
     * @throws IdentityException
     */
    public boolean exists(Role role) throws IdentityException 
    {
        boolean exists = false; 
        try
        {           
            Connection connection = this.getConnection();
            try
            {                       
                String select = "select cn "+
                "from cn="+role.getName().trim()+",ou="+this.roleOu;
                Statement st = connection.createStatement();                
                ResultSet rs = st.executeQuery(select);
                try
                {
                   exists = rs.next();
                }
                finally
                {
                    if(rs!=null)
                    {
                        rs.close();
                    }
                    if(st!=null)
                    {
                        st.close();
                    }
                }
            }
            finally
            {
                if(connection!=null)
                {
                    connection.close();
                }
            }
        }
        catch(SQLException sqe)
        {
            if(sqe.toString().indexOf("No Such Object")!=-1)
            {
                exists = false;
            }
            else
            {
                throw new IdentityException(sqe);
            }
        }
        return exists;
    }
    
    /**
     * Reads role information for the specified role name
     * 
     * @param roleName
     * @return
     * @throws IdentityException
     */
    public Role readRole(String roleName) throws IdentityException
    {
        Role role = null;     
        try
        {           
            Connection connection = this.getConnection();
            try
            {                       
                String select = MessageFormat.format(readRoleSQL,new Object[]{roleName.trim(),this.roleOu});
                Statement st = connection.createStatement();                
                ResultSet rs = st.executeQuery(select);
                try
                {
                   Collection roles = this.readRoles(rs);
                   
                   if(!Tools.isEmpty(roles))
                   {
                       //this should return only one role
                       role = (Role)roles.iterator().next();
                   }
                }
                finally
                {
                    if(rs!=null)
                    {
                        rs.close();
                    }
                    if(st!=null)
                    {
                        st.close();
                    }
                }
            }
            finally
            {
                if(connection!=null)
                {
                    connection.close();
                }
            }
        }
        catch(SQLException sqe)
        {
            if(sqe.toString().indexOf("No Such Object")!=-1)
            {
                role = null;
            }
            else
            {
                throw new IdentityException(sqe);
            }
        }       
        return role;
    }
    
    /**
     * Return all the authorization roles registered in the system. Note: this is a list of all the roles, and not
     * specific to any particular user
     * 
     * @return a Collection of all the authorization roles registered in the system
     * @throws IdentityException
     */
    public Collection readAllRoles() throws IdentityException
    {
       Collection roles = new ArrayList();
       try
       {           
           Connection connection = this.getConnection();
           try
           {                       
               String select = MessageFormat.format(readAllRolesSQL,new Object[]{this.roleOu});
               Statement st = connection.createStatement();                
               ResultSet rs = st.executeQuery(select);
               try
               {
                  roles = this.readRoles(rs);                                    
               }
               finally
               {
                   if(rs!=null)
                   {
                       rs.close();
                   }
                   if(st!=null)
                   {
                       st.close();
                   }
               }
           }
           finally
           {
               if(connection!=null)
               {
                   connection.close();
               }
           }
       }
       catch(SQLException sqe)
       {
           if(sqe.toString().indexOf("No Such Object")!=-1)
           {
               roles = null;
           }
           else
           {
               throw new IdentityException(sqe);
           }
       }
       return roles;
    }
    
    /**
     * Adds a new role to the system
     * 
     * @param role
     * @throws IdentityException
     */
    public void addRole(Role role) throws IdentityException
    {
        //perform validation on the role    
        this.validateRole(role);
        try
        {
            this.storeRole(role);
        }
        catch(Exception e)
        {
            e.printStackTrace();
            throw new IdentityException(e);
        }
    }
    
    /**
     * Removes a role from the system
     * 
     *  @param role
     *  @throws IdentityException
     */
    public void removeRole(Role role) throws IdentityException
    {
        try
        {
            Connection connection = this.getConnection();
            try
            {                    
                String delete = MessageFormat.format(removeRoleSQL,new Object[]{this.roleOu});
                
                PreparedStatement st = connection.prepareStatement(delete);
                st.setString(1,role.getName());                        
                try
                {
                   st.executeUpdate();
                }
                finally
                {                
                    if(st!=null)
                    {
                        st.close();
                    }
                }
            }
            finally
            {
                if(connection!=null)
                {
                    connection.close();
                }
            }
        }
        catch(SQLException sqe)
        {
            throw new IdentityException(sqe);
        }
    }
    
    /**
     * Updates the specified role
     * 
     * @param role
     * @throws IdentityException
     */
    public void updateRole(Role role) throws IdentityException
    {
        //perform corp wide validation of the role    
        this.validateRole(role);
        
        //cleanup any temporary backed up roles
        this.cleanupBackupRole(role);
        
        //make sure that there aren't any backup files existing for this role
        //if they do, the role data for this user needs to be cleaned up
        //while it is in this state, there should be no further changes made to this
        //information
        if(this.isRoleCorrupted(role.getName()))
        {
            throw new IdentityException(role.getName()+" is corrupted. Please cleanup before making any updates on this Role");
        }
                
        //make a temp copy of the entire role that needs to be updated..
        //this copy is used for restoration if there are issues with this update
        Role temp = this.readRole(role.getName());
        String tempRoleName = temp.getName() + ".bak" + "." + System.currentTimeMillis();
        temp.setName(tempRoleName);             
        this.modifyRoleInfo(temp); //write operation
        
        //now perform the actual update of the role in the LDAP server
        try
        {           
            this.removeRole(role); //write operation    
            this.modifyRoleInfo(role); //write operation
        }
        catch(Exception e)
        {
            //restore the original identity from the backed up identity
            //kind of a simulation of rollback          
            if(!this.exists(role))
            {
                //this role has been deleted, we need to restore it from the backed up copy
                Role oldRole = this.readRole(tempRoleName);
                oldRole.setName(role.getName());
                this.modifyRoleInfo(oldRole); //write operation
            }
            throw new IdentityException(e);
        }
        finally
        {               
            this.cleanupBackupRole(role);
        }
    }
    
    
    /**
     * This notification is received from the SSO Engine when the engine performs an automatic logic based 
     * on a trusted assertion received from a partner site. The method can be used by the application to setup its
     * environment in response to the user logged in to the system
     * 
     * @param loginContext
     * @throws IdentityException
     */
    public void processSSOLoginNotification(LoginContext loginContext) throws IdentityException
    {
        //leave this empty....this is a generic built-in module, hence makes no application specific
        //assumptions about setting up its environment
    }    
    //internal private implementation specific stuff--------------------------------------------------------------------------------------------------
    /**
     * 
     * @throws IdentityException
     */
    private void restoreIdentities() throws IdentityException
    {       
        try
        {
            Connection connection = this.getConnection();           
            try
            {
                String select = "select cn "+
                "from ou="+this.identityOu+" where cn=?"; 
                PreparedStatement st = connection.prepareStatement(select);
                st.setString(1,"*.bak.*");
                ResultSet rs = st.executeQuery();
                try
                {
                    while(rs.next())
                    {
                        ParamIndex pIndex = new ParamIndex();
                        String cnBak = rs.getString(pIndex.getIndex());
                        String cn = cnBak.substring(0,cnBak.indexOf(".bak"));
                        if(!this.exists(cn))
                        {
                            //this identity needs to be restored from its backup copy
                            Identity backup = this.read(cnBak);
                            backup.setUserName(cn);
                            this.modifyInfo(backup);
                            this.cleanupBackup(cn);
                        }
                    }
                }
                finally
                {
                    if(rs!=null)
                    {
                        rs.close();
                    }
                    if(st!=null)
                    {
                        st.close();
                    }
                }
            }
            finally
            {
                if(connection!=null)
                {
                    connection.close();
                }
            }
        }
        catch(SQLException sqe)
        {
            throw new IdentityException(sqe);
        }
    }
    
    /**
     * 
     * @throws IdentityException
     */
    private void restoreRoles() throws IdentityException
    {       
        try
        {
            Connection connection = this.getConnection();           
            try
            {
                String select = "select cn "+
                "from ou="+this.roleOu+" where cn=?"; 
                PreparedStatement st = connection.prepareStatement(select);
                st.setString(1,"*.bak.*");
                ResultSet rs = st.executeQuery();
                try
                {
                    while(rs.next())
                    {
                        ParamIndex pIndex = new ParamIndex();
                        String cnBak = rs.getString(pIndex.getIndex());
                        String cn = cnBak.substring(0,cnBak.indexOf(".bak"));
                        Role role = new Role();
                        role.setName(cn);
                        if(!this.exists(role))
                        {
                            //this role needs to be backed up from the back up copy
                            Role backup = this.readRole(cnBak);
                            backup.setName(role.getName());
                            this.modifyRoleInfo(backup);
                            this.cleanupBackupRole(role);
                        }
                    }
                }
                finally
                {
                    if(rs!=null)
                    {
                        rs.close();
                    }
                    if(st!=null)
                    {
                        st.close();
                    }
                }
            }
            finally
            {
                if(connection!=null)
                {
                    connection.close();
                }
            }
        }
        catch(SQLException sqe)
        {
            throw new IdentityException(sqe);
        }
    }
    
    /**
     * 
     * @param userName
     * @return
     * @throws IdentityException
     */
    private boolean isIdentityCorrupted(String userName) throws IdentityException
    {
        boolean corrupted = false;
        try
        {
            Connection connection = this.getConnection();
            try
            {                    
                String select = "select cn "+
                "from ou="+this.identityOu+" where cn=?";
                PreparedStatement st = connection.prepareStatement(select);
                st.setString(1,userName+".bak.*");
                ResultSet rs = st.executeQuery();                        
                try
                {
                    corrupted = rs.next();
                }
                finally
                { 
                    if(rs!=null)
                    {
                        rs.close();
                    }
                    if(st!=null)
                    {
                        st.close();
                    }
                }
            }
            finally
            {
                if(connection!=null)
                {
                    connection.close();
                }
            }
        }
        catch(SQLException sqe)
        {
            throw new IdentityException(sqe);
        }
        return corrupted;
    }
    
    /**
     * 
     * @param userName
     * @return
     * @throws IdentityException
     */
    private boolean isRoleCorrupted(String roleName) throws IdentityException
    {
        boolean corrupted = false;
        try
        {
            Connection connection = this.getConnection();
            try
            {                    
                String select = "select cn "+
                "from ou="+this.roleOu+" where cn=?";
                PreparedStatement st = connection.prepareStatement(select);
                st.setString(1,roleName+".bak.*");
                ResultSet rs = st.executeQuery();                        
                try
                {
                    corrupted = rs.next();
                }
                finally
                { 
                    if(rs!=null)
                    {
                        rs.close();
                    }
                    if(st!=null)
                    {
                        st.close();
                    }
                }
            }
            finally
            {
                if(connection!=null)
                {
                    connection.close();
                }
            }
        }
        catch(SQLException sqe)
        {
            throw new IdentityException(sqe);
        }
        return corrupted;
    }
    
    /**
     * 
     * @param userName
     * @throws IdentityException
     */
    private void cleanupBackup(String userName) throws IdentityException
    {
        //cleanup backed up identities only if this identity actually exists
        if(!this.exists(userName))
        {
            return;
        }
        try
        {
            Connection connection = this.getConnection();
            try
            {                    
                String delete = "delete from ou="+this.identityOu+" where cn=?";
                PreparedStatement st = connection.prepareStatement(delete);
                st.setString(1,userName+".bak.*");                        
                try
                {
                   st.executeUpdate();
                }
                finally
                {                
                    if(st!=null)
                    {
                        st.close();
                    }
                }
            }
            finally
            {
                if(connection!=null)
                {
                    connection.close();
                }
            }
        }
        catch(SQLException sqe)
        {
            throw new IdentityException(sqe);
        }
    }
    
    /**
     * 
     * @param userName
     * @throws IdentityException
     */
    private void cleanupBackupRole(Role role) throws IdentityException
    {
        //cleanup backed up role if role exists
        if(!this.exists(role))
        {
            return;
        }
        try
        {
            Connection connection = this.getConnection();
            try
            {                    
                String delete = "delete from ou="+this.roleOu+" where cn=?";
                PreparedStatement st = connection.prepareStatement(delete);
                st.setString(1,role.getName()+".bak.*");                        
                try
                {
                   st.executeUpdate();
                }
                finally
                {                
                    if(st!=null)
                    {
                        st.close();
                    }
                }
            }
            finally
            {
                if(connection!=null)
                {
                    connection.close();
                }
            }
        }
        catch(SQLException sqe)
        {
            throw new IdentityException(sqe);
        }
    }
    
    /**
     * 
     */ 
    private void modifyInfo(Identity identity) throws IdentityException
    {       
        //perform corp wide validation of the identity  
        this.validateIdentity(identity);
        try
        {
            this.storeIdentity(identity,false);
        }
        catch(Exception e)
        {
            e.printStackTrace();
            throw new IdentityException(e);
        }
    }
    
    /**
     * 
     */ 
    private void modifyRoleInfo(Role role) throws IdentityException
    {                 
        this.validateRole(role);
        try
        {
            this.storeRole(role);
        }
        catch(Exception e)
        {
            e.printStackTrace();
            throw new IdentityException(e);
        }
    }
}
