/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.security.saml;

import java.io.IOException;
import java.io.InputStream;
import java.io.ByteArrayInputStream;
import java.util.Date;
import java.util.Iterator;
import java.util.Collection;
import java.util.ArrayList;
import java.util.Map;
import java.util.Properties;
import java.util.HashMap;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.NameValuePair;
import org.apache.commons.httpclient.contrib.ssl.EasySSLProtocolSocketFactory;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.protocol.Protocol;
import org.opensaml.SAMLAssertion;
import org.opensaml.SAMLAuthenticationQuery;
import org.opensaml.SAMLAuthenticationStatement;
import org.opensaml.SAMLAttributeStatement;
import org.opensaml.SAMLNameIdentifier;
import org.opensaml.SAMLRequest;
import org.opensaml.SAMLResponse;
import org.opensaml.SAMLSubject;
import org.opensaml.SAMLAttribute;
import org.opensaml.SAMLException;
import org.opensaml.provider.SecureRandomIDProvider;

import org.jboss.security.util.Tools;


/**
 * This is an OpenSAML v1.1 based implementation of the SingleSignOn interface. 
 * It produces SSO token in SAML v1.1
 * format.
 * 
 * @author Sohil Shah - sohil.shah@jboss.com
 *
 */
class JBossSingleSignOn implements SingleSignOn 
{
    /**
     * 
     */
    private SecureRandomIDProvider idProvider = null;
    
    /**
     * 
     */
    private Properties properties = null;
    
    
    /**
     * 
     *
     */
    JBossSingleSignOn()
    {
        super();
        idProvider = new SecureRandomIDProvider();
    }
    
    //--------SingleSignOn implementation----------------------------------------------------------------------------
    /**
     * This method is used to set configuration properties on the Single Sign On implementation
     * 
     * @param properties
     */
    public void setProperties(Properties properties)
    {
        this.properties = properties;
    }
    
    /**
     * This method generates a SAML authentication request based on the supplied username and password
     * 
     * @param username
     * @param password
     * @return
     * @throws SSOException
     */
    public String generateAuthRequest(String username,String password) 
    throws SSOException
    {
        try
        {
            String request = null;
            
            //create a SAMLSubject
            SAMLNameIdentifier id = new SAMLNameIdentifier();
            id.setName(username);
            id.setNameQualifier(password);
            id.setFormat(SAMLNameIdentifier.FORMAT_UNSPECIFIED);
            SAMLSubject subject = new SAMLSubject();
            subject.setNameIdentifier(id);      
            SAMLAuthenticationQuery query = new SAMLAuthenticationQuery(subject,
            SAMLAuthenticationStatement.AuthenticationMethod_Password);
            
            SAMLRequest authRequest = new SAMLRequest(query);
            request = authRequest.toString();
            
            return request;
        }
        catch(SAMLException sme)
        {
            throw new SSOException(sme);
        }
    }
    
    /**
     * This method parses a SAML authentication request into a SSOUser domain object
     * 
     * @param request
     * @return
     * @throws SSOException
     */
    public SSOUser parseAuthRequest(String request) throws SSOException
    {
        ByteArrayInputStream bis = null;
        try
        {
            SSOUser user = null;
            
            bis = new ByteArrayInputStream(request.getBytes());
            SAMLRequest authRequest = new SAMLRequest(bis);
            
            SAMLAuthenticationQuery query = (SAMLAuthenticationQuery)authRequest.getQuery();
            SAMLSubject subject = query.getSubject();
                                    
            //get the SAMLNameIdentifier
            SAMLNameIdentifier id = subject.getNameIdentifier();
            String username = id.getName();
            String password = id.getNameQualifier();
            
            user = new SSOUser(username,password);
                        
            return user;
        } 
        catch(SAMLException sme)
        {
            throw new SSOException(sme);
        }
        finally
        {
            if(bis!=null)
            {
                try{bis.close();}catch(IOException e){}
            }
        }
    }
    
    /**
     * This method generates a SAML authentication response based on the supplied username, password, and the 
     * status of the authentication process
     * 
     * @param username
     * @param password
     * @param success
     * @return
     * @throws SSOException
     */
    public String generateAuthResponse(String assertingParty,String username,boolean success) 
    throws SSOException
    {
        ByteArrayInputStream bis = null;
        try
        {
            String response = null;
            
            //construct the SAML Response
            SAMLResponse authResponse = new SAMLResponse();
            authResponse.setId(this.idProvider.getIdentifier());
            
            if(success)
            {
                //create a successfull authenticationstatment                                
                SAMLSubject subject = this.getSubject(username);
                
                //setup the authentication statement for this user
                SAMLAuthenticationStatement authStatement = new SAMLAuthenticationStatement();
                authStatement.setAuthMethod(SAMLAuthenticationStatement.AuthenticationMethod_Password);
                authStatement.setSubject(subject);
                authStatement.setAuthInstant(new Date());
                
                //setup the token secret of this assertion
                String secretSAML = this.getTokenSecret(username);
                bis = new ByteArrayInputStream(secretSAML.getBytes());
                SAMLAttributeStatement st = new SAMLAttributeStatement(bis);
                SAMLAttribute secretAttr = (SAMLAttribute)st.getAttributes().next();
                String secret = (String)secretAttr.getValues().next();
                
                SAMLAttribute tokenSecret = new SAMLAttribute();
                tokenSecret.setName("secret");
                tokenSecret.setNamespace("jbosssso:secret");
                Collection values = new ArrayList();
                values.add(secret);
                tokenSecret.setValues(values);
                SAMLSubject tokenSubject = this.getSubject(username);
                SAMLAttributeStatement secretStatement = new SAMLAttributeStatement();
                secretStatement.addAttribute(tokenSecret);
                secretStatement.setSubject(tokenSubject);
                
                
                
                //create an assertion 
                SAMLAssertion authAssertion = new SAMLAssertion();
                authAssertion.setId(this.idProvider.getIdentifier());
                authAssertion.setIssuer(assertingParty);
                authAssertion.addStatement(authStatement);
                authAssertion.addStatement(secretStatement);
                
                //create the SAMLResponse            
                authResponse.addAssertion(authAssertion);
            }
            else
            {
                SAMLException loginFailed = new SAMLException("login_failed");
                authResponse.setStatus(loginFailed);
            }
            
            response = authResponse.toString();
            
            return response;
        }
        catch(SAMLException sme)
        {
            sme.printStackTrace();
            throw new SSOException(sme);
        }
        finally
        {
            if(bis != null)
            {
                try{bis.close();}catch(Exception e){}
            }
        }
    }
    
    /**
     * This method parses a SAML authentication response and produces an AuthResponse domain object
     * 
     * @param response
     * @return
     * @throws SSOException
     */
    public AuthResponse parseAuthResponse(String resp)
    throws SSOException
    {
        AuthResponse authResponse = null;
        ByteArrayInputStream bis = null;
        boolean success = false;
        String assertToken = null;
        String assertingParty = null;
        String username = null;
        try
        {                        
            bis = new ByteArrayInputStream(resp.getBytes());
            SAMLResponse response = new SAMLResponse(bis);
            
            Iterator assertions = response.getAssertions();
            if(assertions!=null && assertions.hasNext())
            {
                success = true;
                SAMLAssertion authAssertion = (SAMLAssertion)assertions.next();
                assertToken = authAssertion.getId();
                assertingParty = authAssertion.getIssuer();
                
                Iterator statements = authAssertion.getStatements();
                boolean responseValidated = false;
                while(statements.hasNext())
                {
                    Object statement = statements.next();                    
                    if(statement instanceof SAMLAuthenticationStatement)
                    {
                        SAMLAuthenticationStatement authStatement = (SAMLAuthenticationStatement)statement;
                        username = authStatement.getSubject().getNameIdentifier().getName();                
                        SSOUser user = new SSOUser(username,null);
                        authResponse = new AuthResponse(assertingParty,assertToken,user,success);
                    }
                    else if(statement instanceof SAMLAttributeStatement)
                    {
                        SAMLAttributeStatement attributeStatement = (SAMLAttributeStatement)statement;
                        username = attributeStatement.getSubject().getNameIdentifier().getName();
                        SAMLAttribute secretAttribute = (SAMLAttribute)attributeStatement.getAttributes().next();
                        String secret = (String)secretAttribute.getValues().next();
                        
                        responseValidated = this.isSecretValid(username,secret);
                    }
                }
                if(!responseValidated)
                {
                    //validate this here...else throw an exception
                    success = false;
                    if(authResponse == null)
                    {
                        authResponse = new AuthResponse(assertingParty,assertToken,null,success);
                    }
                    else
                    {
                        authResponse = new AuthResponse(authResponse.getAssertingParty(),
                                authResponse.getAssertToken(),
                                authResponse.getUser(),
                                success
                        );
                    }
                }
            }                        
            return authResponse;
        }
        catch(SAMLException sme)
        {
            if(sme.getMessage().equals("login_failed"))
            {
                success = false;
                authResponse = new AuthResponse(assertingParty,assertToken,null,success);
                return authResponse;
            }
            else
            {
                throw new SSOException(sme);
            }
        }
        finally
        {
            if(bis!=null)
            {
                try{bis.close();}catch(Exception e){}
            }
        }
    }
    
    /**
     * This method returns a SAML (xml) representation of a map of partners that are part of
     * this Single Sign On federation
     * 
     * @param partners - a map of partners belonging to the SSO Federation
     * @return
     * @throws SSOException
     */
    public String marshalPartners(Map partners)
    throws SSOException
    {
    	try
    	{
	    	String xml = "";
	    	
	    	if(partners != null && !partners.isEmpty())
	    	{
	    		SAMLAttributeStatement envelope = new SAMLAttributeStatement();
	    		SAMLSubject envelopeSubject = this.getSubject("jbosssso:partnerInfo");
	    		envelope.setSubject(envelopeSubject);
	    		for(Iterator itr=partners.keySet().iterator();itr.hasNext();)
	    		{
	    			String domain = (String)itr.next();
	    			String federationServer = (String)partners.get(domain);
	    			Collection values = new ArrayList();
	    			values.add(federationServer);
	    			
	    			SAMLAttribute attribute = new SAMLAttribute();
	    			attribute.setNamespace("jbosssso:partner");
	    			attribute.setName(domain);
	    			attribute.setValues(values);
	    			
	    			envelope.addAttribute(attribute);
	    		}
	    		xml = envelope.toString();
	    	}
	    	
	    	return xml;
    	}
    	catch(SAMLException sme)
    	{
    		throw new SSOException(sme);
    	}
    }
    
    /**
     * This method unmarshals a SAML (xml) representation of partners that are part of this Single Sign On
     * Federation 
     * It returns a map of these partners with
     * key - domain of the partner
     * value - the address of the federation server of this partner
     * 
     * @return
     * @throws SSOException
     */
    public Map unmarshalPartners(String partners)
    throws SSOException
    {
    	try
    	{
	    	Map partnerMap = new HashMap();
	    	
	    	if(partners != null && partners.trim().length()>0)
	    	{
		    	ByteArrayInputStream bis = new ByteArrayInputStream(partners.getBytes());
		    	SAMLAttributeStatement envelope = new SAMLAttributeStatement(bis);
		    	SAMLSubject subject = envelope.getSubject();
		    	if(!subject.getNameIdentifier().getName().equals("jbosssso:partnerInfo"))
		    	{
		    		throw new RuntimeException(new IllegalAccessException("Invalid Subject"));
		    	}
		    	
		    	for(Iterator itr=envelope.getAttributes();itr.hasNext();)
		    	{
		    		SAMLAttribute attribute = (SAMLAttribute)itr.next();
		    		String domain = attribute.getName();
		    		String federationServer = (String)attribute.getValues().next();
		    		partnerMap.put(domain, federationServer);
		    	}
	    	}
	    	
	    	return partnerMap;
    	}
    	catch(SAMLException sme)
    	{
    		throw new SSOException(sme);
    	}
    }
    //--------SingleSignOn implementation----------------------------------------------------------------------------
    /**
     * 
     */
    private String getTokenSecret(String tokenIdentifier) throws SSOException
    {
        try
        {
            String secret = null;
            
            String url = this.properties.getProperty("trustServer")+"/generateSecret";
            HttpClient client = new HttpClient();
            
            //setup https protocol setting
            Protocol.registerProtocol("https",new Protocol("https",new EasySSLProtocolSocketFactory(),443));
            
            //get a secret for this identifier
            PostMethod post = new PostMethod(url);
            NameValuePair[] data = {new NameValuePair("identifier",tokenIdentifier)};
            post.setRequestBody(data);
            InputStream is = null;
            try
            {
                int statusCode = client.executeMethod(post);
                is = post.getResponseBodyAsStream();
                byte[] response = Tools.getBytes(is);
                secret = new String(response);                                                  
            }
            finally
            {
                if(post != null)
                {
                    post.releaseConnection();
                }
                if(is != null)
                {
                    is.close();
                }
            }
            
            
            return secret;
        }
        catch(Exception e)
        {
            throw new SSOException(e);
        }
    }
    
    /**
     * 
     * @param tokenIdentifier
     * @param secret
     * @return
     */
    private boolean isSecretValid(String tokenIdentifier,String secret) throws SSOException
    {        
        try
        {
            boolean isSecretValid = false;
            
            String url = this.properties.getProperty("trustServer")+"/validateSecret";
            HttpClient client = new HttpClient();
            
            //setup https protocol setting
            Protocol.registerProtocol("https",new Protocol("https",new EasySSLProtocolSocketFactory(),443));
            
            //get a secret for this identifier
            PostMethod post = new PostMethod(url);
            NameValuePair[] data = {new NameValuePair("identifier",tokenIdentifier),new NameValuePair("secret",secret)};
            post.setRequestBody(data);
            InputStream is = null;
            ByteArrayInputStream bis = null;
            try
            {
                int statusCode = client.executeMethod(post);
                is = post.getResponseBodyAsStream();
                byte[] response = Tools.getBytes(is);
                                
                bis = new ByteArrayInputStream(response);
                SAMLAttributeStatement st = new SAMLAttributeStatement(bis);
                SAMLAttribute validAttr = (SAMLAttribute)st.getAttributes().next();
                String validStatus = (String)validAttr.getValues().next();
                
                isSecretValid = (new Boolean(validStatus)).booleanValue();
            }
            finally
            {
                if(post != null)
                {
                    post.releaseConnection();
                }
                if(is != null)
                {
                    is.close();
                }
            }
            
            
            return isSecretValid;
        }
        catch(Exception e)
        {
            throw new SSOException(e);
        }
    }
    
    /**
     * 
     * @param username
     * @return
     */
    private SAMLSubject getSubject(String username) throws SAMLException
    {
       SAMLSubject subject = null;
       
       SAMLNameIdentifier id = new SAMLNameIdentifier();
       id.setName(username);
       id.setFormat(SAMLNameIdentifier.FORMAT_UNSPECIFIED);                
       subject = new SAMLSubject();
       subject.setNameIdentifier(id);
       
       return subject;
    }
}
